from __future__ import absolute_import
from .Interval_core import Interval, get_include
from .LabeledInterval_core import LabeledInterval
from .AIList_core import AIList
from .IntervalArray_core import IntervalArray
from .LabeledIntervalArray_core import LabeledIntervalArray

# This is extracted automatically by the top-level setup.py.
__version__ = '1.0.2'
__author__ = "Kyle S. Smith"


__doc__ = """\
API
======

Basic class
-----------

.. autosummary::
   :toctree: .
   
   Interval
   LabeledInterval
   AIList
   IntervalArray
   LabeledIntervalArray
    
"""